/*
 *      timer.c  --  Reference clock for accurate cycle timing
 *      Copyright (C) 1996  Ted Williams - wa0eir
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *      Versions: 1.1 -  MAR 2003
 */

#include "common.h"
#include "sendCW.h"


/*
 * getCount - gets current value of cycle counter
 */
void getCount (unsigned int *hi, unsigned int *lo)
{  
   __asm__ volatile ("rdtsc; movl %%edx, %0; movl %%eax, %1"
       : "=r" (*hi), "=r" (*lo)
       : /* No input */
       : "%edx", "%eax");
}  


/*
 * timer FUNCTION
 * A waste of good CPU time.
 */
void timer (long usec, float cpu_clock, int rdtsc_ok)
{

   struct timeval tpstart, tpstop;
   struct timezone *tzp = NULL;

   unsigned start_hi;
   unsigned start_lo;
   unsigned now_hi;
   unsigned now_lo;
   int now = 0;
   float fudge_factor = 1.0;

   if (rdtsc_ok == 1)
   {  
      getCount (&start_hi, &start_lo);
      while (now < usec)
      {
         //usleep (usec);
         getCount (&now_hi, &now_lo);
         now = (now_lo - start_lo) / cpu_clock * fudge_factor;
      }
   }
   else                       /* use gettimeofday method */
   {
      if (gettimeofday (&tpstart, tzp))
      {
         perror ("gettimeofday");
         exit (1);
      }

      while (1)
      {
         if (gettimeofday ( &tpstop, tzp))
         {
            perror ("gettimeofday");
            exit (1);
         }
         if (usec < ((tpstop.tv_sec - tpstart.tv_sec) * 1000000
                     + tpstop.tv_usec - tpstart.tv_usec ))
         {
            break;
         }
      }
   }
}


/*
 * Trap function for the probe_clock function
 */
static void sig_ill(int sig)
{
   exit(1);
}


/*
 * probe_clock FUNCTION
 */
int probe_clock (float cpu_clock)
{
   int rdtsc_ok, i;
   unsigned int tmp0, tmp1, tmp2, tmp3;
   pid_t pid;

   switch (pid = fork())
   {
      case -1:
         perror("probe_clock fork");
         exit (1);

      case 0:  /* the child does this */
         signal(SIGILL, sig_ill);
         /* asm(".byte 0xf,0xff"); */  /* test with this illegal instruction */
         __asm__ volatile ("rdtsc" : "=a" (tmp0), "=d" (tmp1));
         __asm__ volatile ("rdtsc" : "=a" (tmp2), "=d" (tmp3));
         signal(SIGILL, SIG_DFL);

         if (tmp0 == tmp2 && tmp1 == tmp3)
         {
            exit(2);
         }
         exit(0);

      default:  /* the parent waits hr for the child to exit */
         if (-1 == wait4(pid, &i, 0, NULL))
            perror("wait4");
      }

   switch (i)
   {
      case 0:

        fprintf (stderr, "\nYour CPU supports the RDTSC instruction. Good!\n"
                          "The CPUspeed factor in Twcw is set to %4.2f.\n"
                          "Let's hope you set that right.\n", cpu_clock);
        rdtsc_ok = 1;
        break;

      case 2:
        fprintf (stderr, "RDTSC present, but does not change!\n");
        rdtsc_ok = 0;
        break;

      default:
        fprintf (stderr, "Illegal instruction exception caught trying RDTSC\n");
        rdtsc_ok = 0;
        break;
   }
   fprintf (stderr, "Using %s for the timebase.\n",
                     rdtsc_ok ? "RDTSC" : "gettimeofday");
   return (rdtsc_ok);
}
